// iMSTemplate.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

/*
 * Uncomment these defines to include the types of interface on which you wish to communicate with the iMS
 */
//#define  ENABLE_SERIAL_PORT_SCAN
//#define  ENABLE_ETHERNET_SCAN
#define  ENABLE_USB_SCAN
#define  IMG_COUNT 30

 /*
 * Uncomment these defines to restrict iMS scan to specific interface ports (using conventional interface type port naming)
 * Note that it is possible to add additional ports by inserting extra lines of "PortMask.push_back(...)" code below
 */
#define SERIAL_PORT_MASK "COM3"
// #define ETHERNET_PORT_MASK  "192.168.2.100"

// All API classes are defined in the iMS namespace.  For convenience, we can declare this here
using namespace iMS;

// Operating Parameters
double Ch1Amp = 60; // Optimum Power, % 
double Ch2Amp = 50; // Optimum Power, %
double Ch3Amp = 60; // Optimum Power, %
double Ch4Amp = 50; // Optimum Power, %
double DDSlvl = 80; // DDS output level, % (DDSPowerLevel)
kHz ImgClk = 90; // Default Image clock rate, KHz,  
std::chrono::nanoseconds delay(300);	// define pulsed SDIO delay 
std::chrono::nanoseconds PW(500);		// define pulsed SDIO pulse width 


auto myiMS = std::make_shared<IMSSystem>();
Image img[IMG_COUNT];

class ImageDownloadSupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_downloading{ true };
	std::atomic<bool> m_error{ false };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ImageDownloadEvents::DOWNLOAD_FINISHED):
			m_downloading.store(false);
			break;
		case (iMS::ImageDownloadEvents::DOWNLOAD_ERROR):
			m_error.store(true);
			break;
		}
	}
	bool Busy() const { return m_downloading.load(); }
	void Reset() { m_downloading.store(true); m_error.store(false); }
	bool Error() const { return m_error.load(); };
};

class ImageVerifySupervisor : public iMS::IEventHandler
{
private:
	std::atomic<bool> m_verifying{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (iMS::ImageDownloadEvents::VERIFY_SUCCESS):
			m_verifying.store(false);
			break;
		case (iMS::ImageDownloadEvents::VERIFY_FAIL):
			m_verifying.store(false);
			break;
		}
	}
	bool Busy() const { return m_verifying.load(); }
	void Reset() { m_verifying.store(true); }
};

class CompensationTableDownloadSupervisor : public IEventHandler
{
private:
	bool m_downloading{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (CompensationEvents::DOWNLOAD_FINISHED): std::cout << "Ch Comp Table Download Finished!" << std::endl; m_downloading = false; break;
		case (CompensationEvents::DOWNLOAD_ERROR): std::cout << "Download error on message " << param << std::endl; break;
		}
	}
	bool Busy() const { return m_downloading; };
};

class CompensationTableVerifySupervisor : public IEventHandler
{
private:
	bool m_verify_success{ false };
	std::atomic<bool> m_verifying{ true };
public:
	void EventAction(void* sender, const int message, const int param)
	{
		switch (message)
		{
		case (CompensationEvents::VERIFY_SUCCESS): std::cout << "Comp Table Verify Successful!" << std::endl; m_verify_success = true; m_verifying.store(false); break;
		case (CompensationEvents::VERIFY_FAIL): std::cout << "Comp Table Verify FAILED!" << std::endl; m_verifying.store(false); break;
		}
	}
	bool Busy() const { return m_verifying.load(); };
	bool Success() const { return m_verify_success; };
};


bool WaitForImageDownload(iMS::IMSSystem& ims, const iMS::Image& img)
{
	bool b_return = true;
	ImageDownload* dl = new ImageDownload(ims, img);
	//	ImageDownload  dl(ims, img);
	ImageDownloadSupervisor ds;

	dl->ImageDownloadEventSubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventSubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (!dl->StartDownload()) {
		dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
		dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);
		return false;
	}

	int timeout = 100;
	while (ds.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
		if (timeout-- == 0)
		{
			b_return = false;
			break;
		}
	}

	dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (ds.Error()) {
		return false;
	}
	if (!timeout) {
		return false;
	}
	if (!b_return)	return false;
	else			return true;
}


bool WaitForCompensationDownload(IMSSystem& ims, const RFChannel& chan, const CompensationTable& tbl)
{
	CompensationTableDownload* dl = new CompensationTableDownload(ims, tbl, chan);
	CompensationTableDownloadSupervisor ds;

	dl->CompensationTableDownloadEventSubscribe(CompensationEvents::DOWNLOAD_FINISHED, &ds);
	dl->CompensationTableDownloadEventSubscribe(CompensationEvents::DOWNLOAD_ERROR, &ds);

	if (!dl->StartDownload()) {
		std::cout << "Error starting download" << std::endl;
		dl->CompensationTableDownloadEventUnsubscribe(CompensationEvents::DOWNLOAD_FINISHED, &ds);
		dl->CompensationTableDownloadEventUnsubscribe(CompensationEvents::DOWNLOAD_ERROR, &ds);
		delete dl;
		return false;
	}

	while (ds.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
	}

	dl->CompensationTableDownloadEventUnsubscribe(CompensationEvents::DOWNLOAD_FINISHED, &ds);
	dl->CompensationTableDownloadEventUnsubscribe(CompensationEvents::DOWNLOAD_ERROR, &ds);

	CompensationTableVerifySupervisor vs;
	dl->CompensationTableDownloadEventSubscribe(CompensationEvents::VERIFY_SUCCESS, &vs);
	dl->CompensationTableDownloadEventSubscribe(CompensationEvents::VERIFY_FAIL, &vs);

	if (!dl->StartVerify()) {
		std::cout << "Error starting verify" << std::endl;
		dl->CompensationTableDownloadEventUnsubscribe(CompensationEvents::VERIFY_SUCCESS, &vs);
		dl->CompensationTableDownloadEventUnsubscribe(CompensationEvents::VERIFY_FAIL, &vs);
		delete dl;
		return false;
	}

	while (vs.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(50));
	}
	if (!vs.Success()) {
		std::cout << "CT Unprogrammed" << std::endl;
	}
	dl->CompensationTableDownloadEventUnsubscribe(CompensationEvents::VERIFY_SUCCESS, &vs);
	dl->CompensationTableDownloadEventUnsubscribe(CompensationEvents::VERIFY_FAIL, &vs);

	int error;
	while ((error = dl->GetVerifyError()) != -1)
	{
		std::cout << "Failure around compensation table point: " << error << std::endl;
	}

	delete dl;
	return true;
}


void Image_Delete()
{
	ImageTableViewer	itv(*myiMS);
	int consumed_memory = 0;
	unsigned itv_size = 0;
	itv_size = itv.Entries();
	for (int i = 0; i < itv_size; i++)
	{
		consumed_memory += itv[i].Size();

	}
	std::cout << "AOD Images Entered: " << itv_size << std::endl;  // itv_size

	for (int i = 0; i < itv_size; i++)
	{
		itv.Erase(itv[0]);
	}
	std::cout << "AOD Images after Erase : " << itv.Entries() << std::endl;
}





static bool wait_for_image_download(IMSSystem& ims, const Image& img)
{
	ImageDownload* dl = new ImageDownload(ims, img);
	ImageDownloadSupervisor ds;

	dl->ImageDownloadEventSubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventSubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (!dl->StartDownload()) {
		std::cout << "Error starting download" << std::endl;
		dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
		dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);
		delete dl;
		return false;
	}

	int timeout = 1000;
	while (ds.Busy()) {
		std::this_thread::sleep_for(std::chrono::milliseconds(5));
		if (timeout-- < 0) {
			std::cout << "Timeout on Image Download." << std::endl;
			return false;
		}
	}

	dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_FINISHED, &ds);
	dl->ImageDownloadEventUnsubscribe(ImageDownloadEvents::DOWNLOAD_ERROR, &ds);

	if (ds.Error()) return false;

	delete dl;
	return true;
}

void downloadImagesToIMS(std::shared_ptr<iMS::IMSSystem> myiMS, const ImageGroup& grp)
{
	int i = 0;
	for (const auto& img : grp) {
		std::cout << "Downloading image " << img.Name() << std::endl;
		if (!wait_for_image_download(*myiMS, img)) {
			std::cout << "Aborting.. <press any key>" << std::endl;
			char c;
			std::cin >> c;
			std::exit(-1);
		}
	}
}


int main(int argc, char* argv[])
{
	//
	// CONNECT TO iMS System
	//
	std::unique_ptr<ConnectionList> connList(new ConnectionList());
	auto& modules = connList->modules();   // returns a ListBase<string>

	// Display all supported modules
	/*for (auto&& mname : modules)
	{
		std::cout << "Module: " << mname << std::endl;
	}*/

	std::cout << "Serial port scan ";
#if !defined(ENABLE_SERIAL_PORT_SCAN)
	std::cout << "disabled" << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_SERIAL") != modules.end()) connList->config("CM_SERIAL").IncludeInScan = false;
#else
	std::cout << "enabled ";
	// default is enabled.
#if defined(SERIAL_PORT_MASK)
	std::cout << "on port " << SERIAL_PORT_MASK << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_SERIAL") != modules.end()) connList->config("CM_SERIAL").PortMask.push_back(SERIAL_PORT_MASK);
#else
	std::cout << "on all available ports" << std::endl;
#endif
#endif

	std::cout << "Ethernet port scan ";
#if !defined(ENABLE_ETHERNET_SCAN)
	std::cout << "disabled" << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_ETH") != modules.end()) connList->config("CM_ETH").IncludeInScan = false;
#else
	std::cout << "enabled ";
	// default is enabled.
#if defined(ETHERNET_PORT_MASK)
	std::cout << "on port " << ETHERNET_PORT_MASK << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_ETH") != modules.end()) connList->config("CM_ETH").PortMask.push_back(ETHERNET_PORT_MASK);
#else
	std::cout << "on all available ports" << std::endl;
#endif
#endif

	std::cout << "USB port scan ";
#if !defined(ENABLE_USB_SCAN)
	std::cout << "disabled" << std::endl;
	if (std::find(modules.begin(), modules.end(), "CM_USBSS") != modules.end()) connList->config("CM_USBSS").IncludeInScan = false;
	if (std::find(modules.begin(), modules.end(), "CM_USBSS") != modules.end()) connList->config("CM_USBLITE").IncludeInScan = false;
#else
	std::cout << "enabled" << std::endl;
#endif

	// Run full system scan
	std::cout << "Scanning for connected iMS Systems. Please wait. . . ." << std::endl;
	std::vector<IMSSystem> fulliMSList = connList->scan();

	auto myiMS = std::make_shared<IMSSystem>();

	if (fulliMSList.empty()) {
		std::cout << "Can't connect to an iMS System.  Press a key to exit." << std::endl;
		while (!_kbhit());
		return -1;
	}
	else {
		std::cout << "Discovered " << fulliMSList.size() << " iMS Systems: " << std::endl;
		int ims_count = 1;
		for (std::vector<IMSSystem>::const_iterator it = fulliMSList.cbegin(); it != fulliMSList.cend(); ++it) {
			std::cout << "  " << ims_count++ << ") " << it->ConnPort() << std::endl;
		}
		while (1) {
			std::cout << "Select a system to continue ... ";
			char ch;
			std::cin >> ch;
			// Don't connect more than 9 systems!
			if ((ch - '0') > 0 && (ch - '0') <= (int)(fulliMSList.size())) {
				*myiMS = fulliMSList[ch - '1'];
				break;
			}
		}
		std::cout << std::endl;

		
	}

	std::cout << "Connecting ... " << std::endl;
	myiMS->Connect();
	std::cout << " Synth FW Version: " << myiMS->Synth().GetVersion() << std::endl;
	std::cout << " Ctrl FW Version: " << myiMS->Ctlr().GetVersion() << std::endl;
	// std::cout << "OK" << std::endl;


	// Application Code goes here
	//
	char ch;
	
	std::cout << " \n Imports existing *.ipp file e.g. generated with Isomet Studio Windows GUI. Internal Clock rate defined in *.ipp" << std::endl;

	std::cout << " For this demo, LUT compensation file can be downloaded and applied. Verify enabled - wait for all 4 Channel LUTs " << std::endl;
	std::cout << " DDS Level: " << DDSlvl << "%, Channel 1..4 Power: " << Ch1Amp << "%/" << Ch2Amp << "%/" << Ch3Amp << "%/" << Ch4Amp << "%" << std::endl;
	std::cout << " Sync-DIO is Pulse Enabled, PW = 0.5usec, Delay 0.3usec " << std::endl;
	
	std::cout << "\nClearing Old Image table \n ";
	ImageTableViewer itv(*myiMS);
	itv.Clear();
	

	SignalPath sp(*myiMS);
	std::cout << "\nDo you want to Load LUT file? [Y/N]...... ";
	std::cin >> ch;
	if (toupper(ch) == 'Y') {
		// Import a channel-scoped LUT here  (place LUT file sin same folder as *.cpp )
		CompensationTableImporter cti("100A-ZPh-ChScpd.LUT");
		if (!cti.IsValid()) {
			std::cout << "/nLUT File is Not Valid!" << std::endl;
			while (!_kbhit());
			return 0;
		}
		else if (cti.IsGlobal()) {
			std::cout << "/nThis example needs a Channel-Scoped LUT.  Yours is Global" << std::endl;
			while (!_kbhit());
			return 0;
		}

		// Retrieve channel LUTs from file and download them to iMS
		CompensationTable channel_lut;
		// std::cout << "\nBe patient !" << std::endl;

		for (int i = RFChannel::min; i <= cti.Channels(); i++) {
			std::cout << "Downloading Channel " << i << " LUT ......verifying" << std::endl;

			// Gets table for channel 'i' from file
			channel_lut = cti.RetrieveChannelLUT(RFChannel(i));

			// Download to iMS
			if (!WaitForCompensationDownload(*myiMS, RFChannel(i), channel_lut)) {
				std::cout << "Aborting.. <press any key>" << std::endl;
				while (!_kbhit());
				return 0;
			}
		}

		// Double check signal path is in channel scope
		if (sp.QueryCompensationScope() == SignalPath::CompensationScope::CHANNEL) {
			std::cout << "Synthesiser set for CHANNEL Compensation Scope" << std::endl;
		}
		else {
			std::cout << "Synthesiser set for GLOBAL Compensation Scope" << std::endl;
		}

	}

	std::cout << "\nDo you want to Enable LUT file? [Y/N]...... ";
	std::cin >> ch;
	if (toupper(ch) == 'N') {
		// Enable compensation
		sp.EnableImagePathCompensation(SignalPath::Compensation::BYPASS, SignalPath::Compensation::BYPASS);
			std::cout << "LUT File Disabled !" << std::endl;
		}
		else {
		sp.EnableImagePathCompensation(SignalPath::Compensation::ACTIVE, SignalPath::Compensation::ACTIVE);
		std::cout << "LUT File Enabled !" << std::endl;
			
		}

	
	// Step 1: Download images from project 
	ImageProject proj;
	if (!proj.Load("6-ImageExample.iip")) {
		;  // Place Image file in same folder as *.cpp 
		std::cout << "Failed to load Image Project" << std::endl;
		exit(-1);
	}
	ImageGroup& grp = proj.FreeImageContainer();  // This is a default group object for all the images in the container that don't belong to named groups

	downloadImagesToIMS(myiMS, grp);
	std::cout << grp.Size() << " Images downloaded." << std::endl;

	// ImageTableViewer itv(*myiMS);
	std::cout << itv;  // Display Image Table

	// Step 2: Set initial conditions
	
	sp.UpdateDDSPowerLevel(Percent(DDSlvl)); // DDS Level

	// Set RF Amplitude to use internal level control (not external modulation)
	sp.SwitchRFAmplitudeControlSource(SignalPath::AmplitudeControl::INDEPENDENT, RFChannel::all);
	// Update per-channel RF Amplitude levels
	sp.UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(Ch1Amp), RFChannel(1));
	sp.UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(Ch2Amp), RFChannel(2));
	sp.UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(Ch3Amp), RFChannel(3));
	sp.UpdateRFAmplitude(SignalPath::AmplitudeControl::INDEPENDENT, Percent(Ch4Amp), RFChannel(4));

	// Enable Synchronous data output on 44way HD connector J7.
	// Assign the active output and the data source
	SignalPath SDOR(*myiMS);
	SDOR.AssignSynchronousOutput(SignalPath::SYNC_SINK::DIG, SignalPath::SYNC_SRC::IMAGE_DIG);
	SDOR.AssignSynchronousOutput(SignalPath::SYNC_SINK::ANLG_A, SignalPath::SYNC_SRC::IMAGE_ANLG_A);
	SDOR.AssignSynchronousOutput(SignalPath::SYNC_SINK::ANLG_B, SignalPath::SYNC_SRC::IMAGE_ANLG_B);
	SDOR.ConfigureSyncDigitalOutput(delay, PW); // enable pulsed SDIO

	// Step 3: Test play images from user input

	std::cout << "\nInternal Image Clock (if selected) = " << ImgClk;
	std::cout << "\nDDSPowerLevel = " << DDSlvl << " %, Ch1/2 WIPER = " << Ch1Amp << " %,  Ch3/4 WIPER = " << Ch3Amp << " %";
	SystemFunc RFGate(*myiMS);
	RFGate.EnableAmplifier(true);
	std::cout << "\n*** PA's enabled ***   -Output Active-\n" << std::endl;


	// select image to play
	std::unique_ptr<ImagePlayer> player = nullptr;

	while (1) {
		std::cout << "Select an image to play (0-" << grp.size() - 1 << ") or 'q' to quit.....";
		std::cin >> ch;
		if (ch == 'q') {
			player->Stop(ImagePlayer::StopStyle::GRACEFULLY);
			itv.Clear();
			break;
		}

		if ((ch < '0') || (ch >= grp.size() + '0')) continue;

		if (player != nullptr) player->Stop(ImagePlayer::StopStyle::IMMEDIATELY);
		player.reset(new ImagePlayer(*myiMS, grp[ch - '0'], ImagePlayer::PlayConfiguration(ImagePlayer::Repeats::FOREVER))); 
		player->cfg.int_ext = ImagePlayer::PointClock::INTERNAL;  // INTERNAL or EXTERNAL, the INT rate recalled from ipp studio project
		player->cfg.trig = ImagePlayer::ImageTrigger::CONTINUOUS;   // EXTERNAL or CONTINUOUS (Internal)
		player->Play(player->cfg.trig);  // simple 'Play()' defaults to Int Trig 
		std::cout << "Playing !" << std::endl;
	}
	
		
		// All done for now.  Set RF power level to minimum. Disconnect from the iMS and tidy up
		// CLEANUP
	std::cout << "\nPress a key to exit..."; 
		while (!_kbhit());
		std::cout << "\nPlease Wait" << std::endl;
		
		RFGate.EnableAmplifier(false);
		sp.SwitchRFAmplitudeControlSource(SignalPath::AmplitudeControl::OFF);
		
		// Work around reset to clear FIFO. Remnents cause problems for subsequent projects w/o dc power cycle of iMS
		SystemFunc sf(*myiMS);
		sf.ConfigureNHF(true, 100, SystemFunc::NHFLocalReset::RESET_ON_COMMS_UNHEALTHY);
		sf.ClearNHF();
		std::this_thread::sleep_for(std::chrono::milliseconds(1000));
	
		myiMS->Disconnect();
		std::cout << "Bye" << std::endl;
		return(0);
	
}